EXPORTS

ipatch_close

_ipatch_code_error
_ipatch_code_errorv
_ipatch_convert_DLS2_init
_ipatch_convert_SF2_init
_ipatch_convert_gig_init

ipatch_file_identify_new
ipatch_file_identify_name
ipatch_close_base_list

ipatch_gerror_message

ipatch_convert_object_to_type_multi_set

;_ipatch_cram_decoder_class_install_properties
;_ipatch_cram_encoder_class_install_properties
;_ipatch_cram_init
;_ipatch_dls2_conn_init
;_ipatch_dls2_info_init
;_ipatch_dls2_sample_init
;_ipatch_iface_prop_init
_ipatch_param_init
;_ipatch_paste_init
_ipatch_sf2_gen_init
;_ipatch_sf2_mod_init
_ipatch_sf2_voice_cache_init_DLS
_ipatch_sf2_voice_cache_init_SF2
_ipatch_sf2_voice_cache_init_gig
_ipatch_type_prop_init
_ipatch_unit_dls_init
_ipatch_unit_generic_init
ipatch_unit_info_new
ipatch_unit_lookup
_ipatch_unit_init
_ipatch_unit_sf2_init

ipatch_application_name DATA
;ipatch_audio_file_get_type

;ipatch_base_find_item_by_midi_locale_ref
ipatch_base_type_get_mime_type
ipatch_base_find_item_by_midi_locale
ipatch_base_find_unused_midi_locale
;ipatch_base_flags_get_type
ipatch_base_get_file_name
ipatch_base_get_type
;ipatch_base_ref_file
ipatch_base_set_file
ipatch_base_set_file_name

ipatch_container_add_unique
ipatch_container_add_connect
ipatch_container_add_disconnect_matched
ipatch_container_append
ipatch_container_count
ipatch_container_get_child_types
ipatch_container_get_children
ipatch_container_get_children_list
ipatch_container_get_type
ipatch_container_get_virtual_types
ipatch_container_init_iter
ipatch_container_insert
ipatch_container_insert_iter
ipatch_container_make_unique
ipatch_container_remove
ipatch_container_remove_all
ipatch_container_remove_connect
ipatch_container_remove_disconnect_matched
ipatch_container_remove_iter
;ipatch_container_set_add_hook
;ipatch_container_set_remove_hook
ipatch_container_type_get_child_types
ipatch_converter_add_input

ipatch_convert_object_to_type
ipatch_convert_objects

ipatch_converter_add_inputs
ipatch_converter_add_output
ipatch_converter_add_outputs
ipatch_converter_convert
ipatch_converter_dls2_inst_to_sf2_voice_cache_get_type
ipatch_converter_dls2_region_to_sf2_voice_cache_get_type
ipatch_converter_dls2_sample_to_sf2_voice_cache_get_type
ipatch_converter_dls2_to_file_get_type
ipatch_converter_file_to_dls2_get_type
ipatch_converter_file_to_gig_get_type
ipatch_converter_file_to_sf2_get_type
;ipatch_converter_find_flags_get_type
;ipatch_converter_flags_get_type
ipatch_converter_get_notes
ipatch_converter_get_type
ipatch_converter_gig_inst_to_sf2_voice_cache_get_type
ipatch_converter_gig_sample_to_sf2_voice_cache_get_type
ipatch_converter_gig_to_file_get_type
ipatch_converter_init
ipatch_converter_log
ipatch_converter_log_next
;ipatch_converter_log_type_get_type
;ipatch_converter_ref_input
;ipatch_converter_ref_inputs
;ipatch_converter_ref_output
;ipatch_converter_ref_outputs
ipatch_converter_reset
;ipatch_converter_sample_file_to_dls2_sample_get_type
;ipatch_converter_sample_file_to_sf2_sample_get_type
ipatch_converter_sf2_inst_to_sf2_voice_cache_get_type
ipatch_converter_sf2_izone_to_sf2_voice_cache_get_type
ipatch_converter_sf2_preset_to_sf2_voice_cache_get_type
ipatch_converter_sf2_pzone_to_sf2_voice_cache_get_type
ipatch_converter_sf2_sample_to_sf2_voice_cache_get_type
ipatch_converter_sf2_to_file_get_type
ipatch_converter_verify
;ipatch_cram_close_ebml_chunk
;ipatch_cram_decoder_converter_get_type
;ipatch_cram_decoder_converter_new
;ipatch_cram_decoder_decode_all
;ipatch_cram_decoder_decode_file
;ipatch_cram_decoder_get_type
;ipatch_cram_decoder_new
;ipatch_cram_decoder_next_file
;ipatch_cram_decoder_set_cram_file
;ipatch_cram_decoder_set_dest_file
;ipatch_cram_decoder_start
;ipatch_cram_dls_file_encode
;ipatch_cram_encoder_converter_get_type
;ipatch_cram_encoder_converter_new
;ipatch_cram_encoder_declare_binary
;ipatch_cram_encoder_declare_file
;ipatch_cram_encoder_declare_flac_audio
;ipatch_cram_encoder_end
;ipatch_cram_encoder_flags_get_type
;ipatch_cram_encoder_get_type
;ipatch_cram_encoder_new
;ipatch_cram_encoder_set_audio_endian
;ipatch_cram_encoder_set_file
;ipatch_cram_encoder_start
;ipatch_cram_encoder_store_audio
;ipatch_cram_encoder_store_audio_interleaved
;ipatch_cram_encoder_store_binary
;ipatch_cram_file_get_type
;ipatch_cram_file_new
;ipatch_cram_fixup_ebml_size
;ipatch_cram_get_ebml_enum
;ipatch_cram_get_ebml_id
;ipatch_cram_read_ebml_chunk
;ipatch_cram_read_varlength
;ipatch_cram_sf2_file_encode
;ipatch_cram_store_ebml_chunk
;ipatch_cram_varlength_decode
;ipatch_cram_varlength_encode
;ipatch_cram_varlength_min_size
;ipatch_cram_varlength_size
;ipatch_cram_write_ebml_chunk
;ipatch_cram_write_ebml_chunk_buf
ipatch_create_converter
;ipatch_create_path
;ipatch_dls2_conn_dest_type_get_type
ipatch_dls2_conn_duplicate
ipatch_dls2_conn_free
ipatch_dls2_conn_get_type
ipatch_dls2_conn_list_duplicate
ipatch_dls2_conn_list_duplicate_fast
ipatch_dls2_conn_list_free
ipatch_dls2_conn_list_set
ipatch_dls2_conn_list_unset
ipatch_dls2_conn_new
;ipatch_dls2_conn_polarity_type_get_type
;ipatch_dls2_conn_src_type_get_type
;ipatch_dls2_conn_transform_masks_get_type
;ipatch_dls2_conn_transform_shifts_get_type
;ipatch_dls2_conn_transform_type_get_type
;ipatch_dls2_find_inst_ref
;ipatch_dls2_find_sample_ref
;ipatch_dls2_flags_get_type
ipatch_dls2_get_info
ipatch_dls2_get_type
ipatch_dls2_info_bag_free
ipatch_dls2_info_bag_new
ipatch_dls2_info_duplicate
ipatch_dls2_info_free
ipatch_dls2_info_get
ipatch_dls2_info_get_property
ipatch_dls2_info_install_class_properties
ipatch_dls2_info_is_defined
ipatch_dls2_info_notify
ipatch_dls2_info_peek
ipatch_dls2_info_set
ipatch_dls2_info_set_property
;ipatch_dls2_info_type_get_type
ipatch_dls2_inst_compare
ipatch_dls2_inst_conn_count
ipatch_dls2_inst_first
;ipatch_dls2_inst_flags_get_type
ipatch_dls2_inst_get_conns
ipatch_dls2_inst_get_info
ipatch_dls2_inst_get_midi_locale
ipatch_dls2_inst_get_type
ipatch_dls2_inst_new
ipatch_dls2_inst_next
ipatch_dls2_inst_set_conn
ipatch_dls2_inst_set_info
ipatch_dls2_inst_set_midi_locale
ipatch_dls2_inst_unset_all_conns
ipatch_dls2_inst_unset_conn
ipatch_dls2_make_unique_name
ipatch_dls2_new
;ipatch_dls2_ref_file
;ipatch_dls2_ref_region_references
ipatch_dls2_region_channel_map_stereo
;ipatch_dls2_region_channel_type_get_type
;ipatch_dls2_region_clear_sample_override_flags
ipatch_dls2_region_conn_count
ipatch_dls2_region_first
;ipatch_dls2_region_flags_get_type
ipatch_dls2_region_get_conns
ipatch_dls2_region_get_info
;ipatch_dls2_region_get_sample_override
ipatch_dls2_region_get_type
ipatch_dls2_region_in_range
ipatch_dls2_region_new
ipatch_dls2_region_next
ipatch_dls2_region_peek_sample
;ipatch_dls2_region_ref_sample
;ipatch_dls2_region_sample_override_get_type
ipatch_dls2_region_set_conn
ipatch_dls2_region_set_info
;ipatch_dls2_region_set_key_range
ipatch_dls2_region_set_sample
;ipatch_dls2_region_set_sample_override
;ipatch_dls2_region_set_sample_override_flags
ipatch_dls2_region_set_velocity_range
ipatch_dls2_region_unset_all_conns
ipatch_dls2_region_unset_conn
;ipatch_dls2_sample_find_store_ref
ipatch_dls2_sample_first
;ipatch_dls2_sample_flags_get_type
;ipatch_dls2_sample_get_size
ipatch_dls2_sample_get_type
ipatch_dls2_sample_info_duplicate
ipatch_dls2_sample_info_free
ipatch_dls2_sample_info_get_property
ipatch_dls2_sample_info_install_class_properties
ipatch_dls2_sample_info_new
ipatch_dls2_sample_info_set_property
ipatch_dls2_sample_new
ipatch_dls2_sample_next
ipatch_dls2_sample_peek_data
;ipatch_dls2_sample_ref_data
ipatch_dls2_sample_set_blank
ipatch_dls2_sample_set_data
ipatch_dls2_set_file
ipatch_dls2_set_info
ipatch_dls_file_get_type
ipatch_dls_file_new
;ipatch_dls_load_connection
;ipatch_dls_load_dlid
;ipatch_dls_load_info
;ipatch_dls_load_pool_table
;ipatch_dls_load_region_header
;ipatch_dls_load_sample_format
;ipatch_dls_load_sample_info
;ipatch_dls_load_wave_link
;ipatch_dls_reader_error_get_type
ipatch_dls_reader_error_quark
;ipatch_dls_reader_fixup
ipatch_dls_reader_get_type
ipatch_dls_reader_load
;ipatch_dls_reader_load_art_list
;ipatch_dls_reader_load_inst_list
;ipatch_dls_reader_load_level_0
;ipatch_dls_reader_load_region_list
;ipatch_dls_reader_load_wave_pool
ipatch_dls_reader_new
;ipatch_dls_reader_set_file
;ipatch_dls_reader_set_pool_table
;ipatch_dls_reader_start
ipatch_dls_writer_get_type
ipatch_dls_writer_new
ipatch_dls_writer_save
;ipatch_dls_writer_set_file
ipatch_dls_writer_set_patch
ipatch_dump_object
;ipatch_error_get_type
ipatch_error_quark
ipatch_file_assign_fd
ipatch_file_buf_commit
;ipatch_file_buf_free
;ipatch_file_buf_init
ipatch_file_buf_load
ipatch_file_buf_memset
ipatch_file_buf_read
ipatch_file_buf_read_s16
ipatch_file_buf_read_s32
ipatch_file_buf_read_s64
ipatch_file_buf_read_s8
ipatch_file_buf_read_u16
ipatch_file_buf_read_u32
ipatch_file_buf_read_u64
ipatch_file_buf_read_u8
ipatch_file_buf_seek
ipatch_file_buf_write
ipatch_file_buf_write_s16
ipatch_file_buf_write_s32
ipatch_file_buf_write_s64
ipatch_file_buf_write_s8
ipatch_file_buf_write_u16
ipatch_file_buf_write_u32
ipatch_file_buf_write_u64
ipatch_file_buf_write_u8
ipatch_file_close
;ipatch_file_convert_type
ipatch_file_default_close_method
ipatch_file_default_get_size_method
ipatch_file_default_getfd_method
ipatch_file_default_open_method
ipatch_file_default_read_method
ipatch_file_default_seek_method
ipatch_file_default_write_method
ipatch_file_get_fd
ipatch_file_get_iofuncs
ipatch_file_get_name
ipatch_file_get_position
ipatch_file_get_size
ipatch_file_get_type
ipatch_file_identify
ipatch_file_identify_open
;ipatch_file_is_open
ipatch_file_new
ipatch_file_open
;ipatch_file_open_with_fd
;ipatch_file_open_with_io_channel
ipatch_file_read
ipatch_file_read_eof
ipatch_file_read_s16
ipatch_file_read_s32
ipatch_file_read_s64
ipatch_file_read_s8
ipatch_file_read_u16
ipatch_file_read_u32
ipatch_file_read_u64
ipatch_file_read_u8
;ipatch_file_ref_io_channel
ipatch_file_seek
ipatch_file_seek_eof
ipatch_file_set_big_endian
ipatch_file_set_iofuncs
ipatch_file_set_iofuncs_null
ipatch_file_set_iofuncs_static
ipatch_file_set_little_endian
ipatch_file_set_name
ipatch_file_update_position
ipatch_file_write
ipatch_file_write_s16
ipatch_file_write_s32
ipatch_file_write_s64
ipatch_file_write_s8
ipatch_file_write_u16
ipatch_file_write_u32
ipatch_file_write_u64
ipatch_file_write_u8
ipatch_find_converter
;ipatch_find_paste_handler
ipatch_glist_unref_free
ipatch_gerror_message
;ipatch_get_conversion_dest_types
;ipatch_get_conversion_src_types
;ipatch_get_conversion_table
;ipatch_get_paste_dest_types
;ipatch_get_paste_src_types
;ipatch_gig_control_type_get_type
ipatch_gig_dimension_first
ipatch_gig_dimension_get_type
ipatch_gig_dimension_new
ipatch_gig_dimension_next
;ipatch_gig_dimension_type_get_type
ipatch_gig_effects_init
ipatch_gig_effects_to_gen_array
ipatch_gig_file_get_type
ipatch_gig_file_new
;ipatch_gig_filter_type_get_type
ipatch_gig_get_type
ipatch_gig_inst_first
ipatch_gig_inst_get_type
ipatch_gig_inst_new
ipatch_gig_inst_next
;ipatch_gig_load_dimension_info
;ipatch_gig_load_dimension_names
;ipatch_gig_load_group_names
;ipatch_gig_load_region_header
;ipatch_gig_load_sample_info
ipatch_gig_new
ipatch_gig_parse_effects
;ipatch_gig_reader_load_inst_lart
;ipatch_gig_reader_load_region_list
;ipatch_gig_reader_load_sub_regions
ipatch_gig_region_first
;ipatch_gig_region_flags_get_type
ipatch_gig_region_get_type
ipatch_gig_region_new
ipatch_gig_region_new_dimension
ipatch_gig_region_next
ipatch_gig_region_remove_dimension
;ipatch_gig_region_set_key_range
ipatch_gig_region_set_velocity_range
ipatch_gig_sample_first
ipatch_gig_sample_get_type
ipatch_gig_sample_new
ipatch_gig_sample_next
ipatch_gig_store_effects
ipatch_gig_sub_region_first
;ipatch_gig_sub_region_flags_get_type
ipatch_gig_sub_region_get_type
ipatch_gig_sub_region_new
ipatch_gig_sub_region_next
;ipatch_gig_sub_region_ref_sample
ipatch_gig_sub_region_set_sample
;ipatch_gig_sub_region_set_sample_info_override
ipatch_gig_to_sf2_timecents
ipatch_gig_volsust_to_sf2_centibels
;ipatch_iface_find_property
;ipatch_iface_get
;ipatch_iface_get_property
;ipatch_iface_get_valist
;ipatch_iface_install_property
;ipatch_iface_install_property_handlers
;ipatch_iface_list_properties
;ipatch_iface_set
;ipatch_iface_set_property
;ipatch_iface_set_valist
ipatch_init
ipatch_item_changed
;ipatch_item_clear_flags
ipatch_item_copy
;ipatch_item_describe
ipatch_item_duplicate
ipatch_item_first
;ipatch_item_flags_get_type
ipatch_item_get_atomic
ipatch_item_get_base
ipatch_item_get_flags
ipatch_item_get_type
ipatch_item_next
;ipatch_item_peek_ancestor_type
ipatch_item_peek_ancestor_by_type
ipatch_item_peek_base
ipatch_item_peek_parent
ipatch_item_prop_connect
ipatch_item_prop_disconnect
ipatch_item_prop_notify
ipatch_item_prop_notify_by_name

;ipatch_item_pspec_title DATA
; Getter function returning title property GParamSpec as a convenience to derived type.
ipatch_item_get_pspec_title

;ipatch_item_ref_ancestor_type
;ipatch_item_ref_base
;ipatch_item_ref_parent
ipatch_item_remove
ipatch_item_set_atomic
ipatch_item_set_flags
ipatch_item_get_parent
ipatch_item_set_parent
;ipatch_item_set_prop_notify_hook
ipatch_item_unparent
ipatch_iter_GList_count
ipatch_iter_GList_first
ipatch_iter_GList_get
ipatch_iter_GList_index
ipatch_iter_GList_init
ipatch_iter_GList_insert
ipatch_iter_GList_last
;ipatch_iter_GList_methods DATA
ipatch_iter_GList_next
ipatch_iter_GList_remove
ipatch_iter_GSList_count
ipatch_iter_GSList_first
ipatch_iter_GSList_get
ipatch_iter_GSList_index
ipatch_iter_GSList_init
ipatch_iter_GSList_insert
ipatch_iter_GSList_last
;ipatch_iter_GSList_methods DATA
ipatch_iter_GSList_next
ipatch_iter_GSList_remove
ipatch_iter_alloc
ipatch_iter_array_count
ipatch_iter_array_first
ipatch_iter_array_get
ipatch_iter_array_index
ipatch_iter_array_init
ipatch_iter_array_insert
ipatch_iter_array_last
;ipatch_iter_array_methods DATA
ipatch_iter_array_next
ipatch_iter_array_remove
ipatch_iter_duplicate
ipatch_iter_free
ipatch_iter_get_type
ipatch_list_duplicate
ipatch_list_get_type
ipatch_list_init_iter
ipatch_list_new
ipatch_lookup_converter_info
;ipatch_marshal_ENUM__FLAGS
ipatch_md5_final
ipatch_md5_init
ipatch_md5_update
ipatch_param_find_property
ipatch_param_get
ipatch_param_get_property
ipatch_param_get_valist
ipatch_param_install_property
ipatch_param_list_properties
ipatch_param_set
ipatch_param_set_property
ipatch_param_set_valist
ipatch_param_spec_range
ipatch_param_spec_range_get_type

ipatch_is_paste_possible

;ipatch_paste_choice_flags_get_type
;ipatch_paste_choice_get_type
;ipatch_paste_default_handler
;ipatch_paste_execute
;ipatch_paste_find_flags_get_type

ipatch_paste_get_add_list
;ipatch_paste_get_choice
ipatch_paste_get_type
;ipatch_paste_is_possible
ipatch_paste_new
ipatch_paste_objects
;ipatch_paste_set_objects
ipatch_paste_finish

ipatch_range_copy
ipatch_range_free
ipatch_range_get_type
ipatch_range_new
;ipatch_register_converter
ipatch_register_paste_handler
;ipatch_riff_chunk_type_get_type
ipatch_riff_close_chunk
ipatch_riff_get_chunk
ipatch_riff_get_chunk_array
ipatch_riff_get_chunk_level
ipatch_riff_get_error
;ipatch_riff_get_file
ipatch_riff_get_position
ipatch_riff_get_total_size
ipatch_riff_message_detail
;ipatch_riff_mode_get_type
;ipatch_riff_parser_error_get_type
;ipatch_riff_parser_error_quark
;ipatch_riff_parser_get_type
;ipatch_riff_parser_new
ipatch_riff_new
ipatch_riff_pop_state
ipatch_riff_push_state
ipatch_riff_read_chunk
ipatch_riff_read_chunk_verify
;ipatch_riff_set_file
ipatch_riff_skip_chunks
ipatch_riff_start_read
ipatch_riff_start_read_chunk
;ipatch_riff_status_get_type
ipatch_riff_write_chunk
;ipatch_sample_channel_get_type
;ipatch_sample_channel_route_get_type
;ipatch_sample_data_find_flags_get_type
;ipatch_sample_data_find_store_ref
;ipatch_sample_data_first

ipatch_sample_data_get_cache_sample

ipatch_set_sample_store_swap_file_name
ipatch_base_save_to_filename
ipatch_compact_sample_store_swap
ipatch_set_sample_store_swap_max_memory
ipatch_get_sample_store_swap_unused_size

;ipatch_sample_data_get_list

;ipatch_sample_data_get_primary_channels
;ipatch_sample_data_get_primary_format
;ipatch_sample_data_get_primary_frame_size
ipatch_sample_data_get_size
ipatch_sample_data_get_type
ipatch_sample_data_get_native_format
ipatch_sample_data_new
;ipatch_sample_data_next
;ipatch_sample_data_ref_blank
;ipatch_sample_data_ref_primary_store
;ipatch_sample_data_set_size
;ipatch_sample_endian_get_type
;ipatch_sample_file_get_save_types
;ipatch_sample_file_get_type
;ipatch_sample_find_flags_get_type
;ipatch_sample_find_source
ipatch_sample_format_verify
ipatch_sample_get_format
ipatch_sample_get_type
ipatch_sample_get_loop_types
ipatch_sample_handle_close
ipatch_sample_handle_get_max_frames
ipatch_sample_handle_open
ipatch_sample_handle_read
ipatch_sample_get_sample_data
ipatch_sample_install_property
ipatch_sample_install_property_readonly
;ipatch_sample_loop_type_get_type
ipatch_sample_new_property_param_spec
;ipatch_sample_new_transform_func_array
ipatch_sample_data_open_cache_sample
;ipatch_sample_sign_get_type
;ipatch_sample_source_close
;ipatch_sample_source_get_format
;ipatch_sample_source_get_frame_size
;ipatch_sample_source_get_native_format
;ipatch_sample_source_get_transform
;ipatch_sample_source_read
;ipatch_sample_source_set_format
;ipatch_sample_speed_get_type


ipatch_sample_cache_clean

ipatch_sample_save_to_file

;ipatch_sample_store_FALSE
;ipatch_sample_store_RAM_get_location
;ipatch_sample_store_RAM_get_type
;ipatch_sample_store_RAM_new
;ipatch_sample_store_RAM_set_location
;ipatch_sample_store_ROM_get_location
;ipatch_sample_store_ROM_get_type
;ipatch_sample_store_ROM_new
;ipatch_sample_store_ROM_set_location
;ipatch_sample_store_TRUE
;ipatch_sample_store_activate
;ipatch_sample_store_alloc
;ipatch_sample_store_copy
ipatch_sample_store_cache_get_type
;ipatch_sample_store_duplicate
;ipatch_sample_store_file_get_location

ipatch_sample_store_cache_open
ipatch_sample_store_cache_close

ipatch_sample_store_file_get_type
ipatch_sample_store_file_new
;ipatch_sample_store_file_peek_file
;ipatch_sample_store_file_ref_file
;ipatch_sample_store_file_set_file
;ipatch_sample_store_file_set_location
ipatch_sample_store_first
;ipatch_sample_store_flags_get_type
;ipatch_sample_store_get_active
;ipatch_sample_store_get_format
;ipatch_sample_store_get_frame_size
;ipatch_sample_store_get_size
;ipatch_sample_store_get_speed
ipatch_sample_store_get_type
;ipatch_sample_store_is_readable
;ipatch_sample_store_is_writable
ipatch_sample_store_next
;ipatch_sample_store_read
;ipatch_sample_store_set_format
ipatch_sample_store_swap_get_type
ipatch_sample_store_swap_new
;ipatch_sample_store_transform_init
;ipatch_sample_store_transform_init_size
;ipatch_sample_store_transform_new
;ipatch_sample_store_transform_new_size
;ipatch_sample_store_transform_read
;ipatch_sample_store_transform_write
;ipatch_sample_store_write
ipatch_sample_transform_alloc
ipatch_sample_transform_alloc_size
ipatch_sample_transform_convert
ipatch_sample_transform_free_buffers
ipatch_sample_transform_get_buffers
ipatch_sample_transform_get_frame_sizes
ipatch_sample_transform_get_max_frames
;ipatch_sample_transform_get_type
ipatch_sample_transform_new
;ipatch_sample_transform_set_buffers
ipatch_sample_transform_set_formats
;ipatch_sample_width_get_type
;ipatch_sample_width_sizes DATA
ipatch_set_application_name

ipatch_sf2_file_get_sample_pos
ipatch_sf2_file_get_sample_size
ipatch_sf2_file_get_type
ipatch_sf2_file_new
ipatch_sf2_file_set_sample_pos
ipatch_sf2_file_set_sample_size
;ipatch_sf2_find_inst_ref
;ipatch_sf2_find_preset_ref
;ipatch_sf2_find_sample_ref
ipatch_sf2_free_info_array
;ipatch_sf2_gen_abs_array DATA
;ipatch_sf2_gen_abs_valid_mask DATA
;ipatch_sf2_gen_add_mask DATA
ipatch_sf2_gen_amount_to_value
ipatch_sf2_gen_array_duplicate
ipatch_sf2_gen_array_free
ipatch_sf2_gen_array_get_type
ipatch_sf2_gen_array_init
ipatch_sf2_gen_array_new
ipatch_sf2_gen_array_offset

;ipatch_sf2_gen_info DATA
; Getter function returning ipatch_sf2_gen_info table.
ipatch_sf2_get_gen_info

ipatch_sf2_gen_item_get_type
ipatch_sf2_gen_item_copy_all

;ipatch_sf2_gen_ofs_array DATA
;ipatch_sf2_gen_ofs_valid_mask DATA
;ipatch_sf2_gen_sample_modes_get_type
;ipatch_sf2_gen_type_get_type
;ipatch_sf2_genid_default_value
;ipatch_sf2_genid_offset
;ipatch_sf2_genid_unit_str
ipatch_sf2_get_info
ipatch_sf2_get_info_array
ipatch_sf2_get_info_max_size
ipatch_sf2_get_type
ipatch_sf2_info_id_is_valid
;ipatch_sf2_info_type_get_type
ipatch_sf2_inst_first
ipatch_sf2_inst_get_name
ipatch_sf2_inst_get_type
;ipatch_sf2_inst_has_global_zone
ipatch_sf2_inst_new
ipatch_sf2_inst_new_zone
ipatch_sf2_inst_next
ipatch_sf2_inst_set_name
ipatch_sf2_izone_first
ipatch_sf2_izone_get_type
ipatch_sf2_izone_new
ipatch_sf2_izone_next
;ipatch_sf2_izone_ref_sample
ipatch_sf2_izone_set_sample
;ipatch_sf2_load_bag
;ipatch_sf2_load_gen
;ipatch_sf2_load_ihdr
;ipatch_sf2_load_mod
;ipatch_sf2_load_phdr
;ipatch_sf2_load_shdr
ipatch_sf2_make_unique_name
;ipatch_sf2_mod_control_get_type
;ipatch_sf2_mod_control_palette_get_type
;ipatch_sf2_mod_direction_get_type
ipatch_sf2_mod_duplicate
;ipatch_sf2_mod_field_masks_get_type
;ipatch_sf2_mod_field_shifts_get_type
ipatch_sf2_mod_free
ipatch_sf2_mod_get_type
ipatch_sf2_mod_list_duplicate

ipatch_sf2_mod_list_get_type

ipatch_sf2_mod_list_free
ipatch_sf2_mod_list_change
ipatch_sf2_mod_list_get_default
ipatch_sf2_mod_list_offset
ipatch_sf2_mod_list_override
ipatch_sf2_mod_list_insert
ipatch_sf2_mod_list_remove
ipatch_sf2_mod_new
ipatch_sf2_mod_item_get_type
;ipatch_sf2_mod_polarity_get_type
;ipatch_sf2_mod_transform_get_type
;ipatch_sf2_mod_type_get_type
ipatch_sf2_new
ipatch_sf2_preset_compare
ipatch_sf2_preset_first
ipatch_sf2_preset_get_midi_locale
ipatch_sf2_preset_get_name
ipatch_sf2_preset_get_type
;ipatch_sf2_preset_has_global_zone
ipatch_sf2_preset_new
ipatch_sf2_preset_new_zone
ipatch_sf2_preset_next
ipatch_sf2_preset_set_midi_locale
ipatch_sf2_preset_set_name
ipatch_sf2_pzone_first
ipatch_sf2_pzone_get_type
ipatch_sf2_pzone_new
ipatch_sf2_pzone_next
;ipatch_sf2_pzone_ref_inst
ipatch_sf2_pzone_set_inst
ipatch_sf2_reader_get_type
ipatch_sf2_reader_load
ipatch_sf2_reader_new
;ipatch_sf2_reader_set_file
;ipatch_sf2_ref_file
;ipatch_sf2_ref_zone_references
;ipatch_sf2_sample_find_store_ref
ipatch_sf2_sample_first
;ipatch_sf2_sample_flags_get_type
ipatch_sf2_sample_get_name
;ipatch_sf2_sample_get_size
ipatch_sf2_sample_get_type
ipatch_sf2_sample_new
ipatch_sf2_sample_next
ipatch_sf2_sample_peek_data
ipatch_sf2_sample_peek_linked
;ipatch_sf2_sample_ref_data
;ipatch_sf2_sample_ref_linked
ipatch_sf2_sample_set_blank
ipatch_sf2_sample_set_data
ipatch_sf2_sample_set_linked
ipatch_sf2_sample_set_name
ipatch_sf2_set_file
ipatch_sf2_set_info
;ipatch_sf2_units_clamp
;ipatch_sf2_units_range_intersect
;ipatch_sf2_units_sfont_to_user
;ipatch_sf2_units_sfont_to_user_str
;ipatch_sf2_units_user_to_sfont
ipatch_sf2_voice_cache_add_voice
ipatch_sf2_voice_cache_get_type
ipatch_sf2_voice_cache_new
ipatch_sf2_voice_cache_optimize
ipatch_sf2_voice_cache_select
ipatch_sf2_voice_cache_set_voice_range
ipatch_sf2_voice_cache_set_default_mods


ipatch_sf2_voice_cache_sample_data
ipatch_sf2_voice_cache_update

;ipatch_sf2_voice_convert_sample_data
;ipatch_sf2_voice_sel_type_get_type
ipatch_sf2_write_bag
ipatch_sf2_write_gen
ipatch_sf2_write_ihdr
ipatch_sf2_write_mod
ipatch_sf2_write_phdr
ipatch_sf2_write_shdr
ipatch_sf2_writer_get_type
ipatch_sf2_writer_new
ipatch_sf2_writer_save
;ipatch_sf2_writer_set_file
ipatch_sf2_writer_set_patch
;ipatch_sf2_zone_class_install_gen_properties
;ipatch_sf2_zone_copy_gen_array
;ipatch_sf2_zone_fill_gen_array
ipatch_sf2_zone_first
;ipatch_sf2_zone_gen_count
;ipatch_sf2_zone_genid_get_prop_name
;ipatch_sf2_zone_genid_get_pspec
;ipatch_sf2_zone_get_gen
;ipatch_sf2_zone_get_gen_property
;ipatch_sf2_zone_get_mods
ipatch_sf2_zone_get_type
;ipatch_sf2_zone_in_range
;ipatch_sf2_zone_insert_mod
;ipatch_sf2_zone_mod_count
ipatch_sf2_zone_next
;ipatch_sf2_zone_peek_item
;ipatch_sf2_zone_ref_item
;ipatch_sf2_zone_remove_mod
;ipatch_sf2_zone_set_gen
;ipatch_sf2_zone_set_gen_property
;ipatch_sf2_zone_set_item
;ipatch_sf2_zone_set_key_range
;ipatch_sf2_zone_set_mod
;ipatch_sf2_zone_set_velocity_range
;ipatch_sf2_zone_unset_gen

ipatch_strconcat_num

ipatch_sli_get_type
;ipatch_sli_inst_cat_strings
;getter fonction that returns pointer on ipatch_sli_inst_cat_strings table
ipatch_sli_inst_get_cat_strings
ipatch_sli_sample_get_type
;ipatch_sli_inst_cat_map
;getter fonction that returns pointer on ipatch_sli_inst_cat_map table
ipatch_sli_inst_get_cat_map
ipatch_sli_inst_get_category_as_path
ipatch_sli_inst_get_type
ipatch_sli_zone_get_type

ipatch_type_find_property
ipatch_type_find_types_with_property
ipatch_type_get
ipatch_type_get_dynamic_func
ipatch_type_get_property
ipatch_type_get_valist
ipatch_type_install_property
ipatch_type_list_properties
ipatch_type_object_get
ipatch_type_object_get_property
ipatch_type_object_get_valist
;ipatch_type_register_func
ipatch_type_set
ipatch_type_set_property
ipatch_type_set_valist
;ipatch_unit_centibel_to_decibel
;ipatch_unit_centibel_to_dls_gain
ipatch_unit_cents_to_hertz
ipatch_unit_class_lookup_map
ipatch_unit_class_register_map
;ipatch_unit_class_type_get_type
ipatch_unit_conversion_lookup
ipatch_unit_conversion_register
ipatch_unit_convert
;ipatch_unit_decibel_to_centibel
;ipatch_unit_decibel_to_dls_gain
ipatch_unit_dls_abs_pitch_to_hertz
ipatch_unit_dls_abs_pitch_to_sf2_abs_pitch
ipatch_unit_dls_abs_time_to_seconds
ipatch_unit_dls_abs_time_to_sf2_abs_time
ipatch_unit_dls_class_convert
;ipatch_unit_dls_gain_to_centibel
;ipatch_unit_dls_gain_to_decibel
;ipatch_unit_flags_get_type
ipatch_unit_hertz_to_cents
ipatch_unit_hertz_to_dls_abs_pitch
ipatch_unit_hertz_to_sf2_abs_pitch
;ipatch_unit_info_alloc
ipatch_unit_info_free
;ipatch_unit_lookup_by_id
ipatch_unit_lookup_by_name
ipatch_unit_percent_to_tenth_percent
ipatch_unit_register
ipatch_unit_seconds_to_dls_abs_time
ipatch_unit_seconds_to_sf2_abs_time
ipatch_unit_sf2_abs_pitch_to_dls_abs_pitch
ipatch_unit_sf2_abs_pitch_to_hertz
ipatch_unit_sf2_abs_time_to_dls_abs_time
ipatch_unit_sf2_abs_time_to_seconds
ipatch_unit_tenth_percent_to_percent
;ipatch_unit_type_get_type
ipatch_unit_user_class_convert
ipatch_value_get_range
ipatch_value_set_range
ipatch_value_set_static_range
ipatch_version
ipatch_vbank_get_type
ipatch_vbank_new
ipatch_virtual_container_get_type
ipatch_virtual_dls2_melodic_get_type
ipatch_virtual_dls2_percussion_get_type
ipatch_virtual_dls2_samples_get_type
ipatch_virtual_gig_melodic_get_type
ipatch_virtual_gig_percussion_get_type
ipatch_virtual_gig_samples_get_type
ipatch_virtual_sf2_inst_get_type
ipatch_virtual_sf2_melodic_get_type
ipatch_virtual_sf2_percussion_get_type
ipatch_virtual_sf2_rom_get_type
ipatch_virtual_sf2_samples_get_type

ipatch_xml_copy
ipatch_xml_decode_property
ipatch_xml_decode_object
ipatch_xml_destroy
ipatch_xml_encode_object
ipatch_xml_get_name
ipatch_xml_get_attribute
ipatch_xml_new_node
ipatch_xml_to_str
ipatch_xml_test_attribute
ipatch_xml_test_name
ipatch_xml_load_from_file

ipatch_simple_paste

ipatch_snd_file_get_type
