
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __IPATCH_BUILTIN_ENUMS_H__
#define __IPATCH_BUILTIN_ENUMS_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "IpatchBase.h" */
GType ipatch_base_flags_get_type (void);
#define IPATCH_TYPE_BASE_FLAGS (ipatch_base_flags_get_type())
/* enumerations from "IpatchConverter.h" */
GType ipatch_converter_log_type_get_type (void);
#define IPATCH_TYPE_CONVERTER_LOG_TYPE (ipatch_converter_log_type_get_type())
GType ipatch_converter_count_get_type (void);
#define IPATCH_TYPE_CONVERTER_COUNT (ipatch_converter_count_get_type())
GType ipatch_converter_flags_get_type (void);
#define IPATCH_TYPE_CONVERTER_FLAGS (ipatch_converter_flags_get_type())
GType ipatch_converter_priority_get_type (void);
#define IPATCH_TYPE_CONVERTER_PRIORITY (ipatch_converter_priority_get_type())
/* enumerations from "IpatchDLS2.h" */
GType ipatch_dls2_flags_get_type (void);
#define IPATCH_TYPE_DLS2_FLAGS (ipatch_dls2_flags_get_type())
/* enumerations from "IpatchDLS2Conn.h" */
GType ipatch_dls2_conn_src_type_get_type (void);
#define IPATCH_TYPE_DLS2_CONN_SRC_TYPE (ipatch_dls2_conn_src_type_get_type())
GType ipatch_dls2_conn_dest_type_get_type (void);
#define IPATCH_TYPE_DLS2_CONN_DEST_TYPE (ipatch_dls2_conn_dest_type_get_type())
GType ipatch_dls2_conn_transform_type_get_type (void);
#define IPATCH_TYPE_DLS2_CONN_TRANSFORM_TYPE (ipatch_dls2_conn_transform_type_get_type())
GType ipatch_dls2_conn_polarity_type_get_type (void);
#define IPATCH_TYPE_DLS2_CONN_POLARITY_TYPE (ipatch_dls2_conn_polarity_type_get_type())
GType ipatch_dls2_conn_transform_masks_get_type (void);
#define IPATCH_TYPE_DLS2_CONN_TRANSFORM_MASKS (ipatch_dls2_conn_transform_masks_get_type())
GType ipatch_dls2_conn_transform_shifts_get_type (void);
#define IPATCH_TYPE_DLS2_CONN_TRANSFORM_SHIFTS (ipatch_dls2_conn_transform_shifts_get_type())
/* enumerations from "IpatchDLS2Info.h" */
GType ipatch_dls2_info_type_get_type (void);
#define IPATCH_TYPE_DLS2_INFO_TYPE (ipatch_dls2_info_type_get_type())
/* enumerations from "IpatchDLS2Inst.h" */
GType ipatch_dls2_inst_flags_get_type (void);
#define IPATCH_TYPE_DLS2_INST_FLAGS (ipatch_dls2_inst_flags_get_type())
/* enumerations from "IpatchDLS2Region.h" */
GType ipatch_dls2_param_get_type (void);
#define IPATCH_TYPE_DLS2_PARAM (ipatch_dls2_param_get_type())
GType ipatch_dls2_region_channel_type_get_type (void);
#define IPATCH_TYPE_DLS2_REGION_CHANNEL_TYPE (ipatch_dls2_region_channel_type_get_type())
GType ipatch_dls2_region_flags_get_type (void);
#define IPATCH_TYPE_DLS2_REGION_FLAGS (ipatch_dls2_region_flags_get_type())
/* enumerations from "IpatchDLS2Sample.h" */
GType ipatch_dls2_sample_flags_get_type (void);
#define IPATCH_TYPE_DLS2_SAMPLE_FLAGS (ipatch_dls2_sample_flags_get_type())
/* enumerations from "IpatchDLSReader.h" */
GType ipatch_dls_reader_error_get_type (void);
#define IPATCH_TYPE_DLS_READER_ERROR (ipatch_dls_reader_error_get_type())
/* enumerations from "IpatchFile.h" */
GType ipatch_file_flags_get_type (void);
#define IPATCH_TYPE_FILE_FLAGS (ipatch_file_flags_get_type())
GType ipatch_file_identify_order_get_type (void);
#define IPATCH_TYPE_FILE_IDENTIFY_ORDER (ipatch_file_identify_order_get_type())
/* enumerations from "IpatchGigDimension.h" */
GType ipatch_gig_dimension_type_get_type (void);
#define IPATCH_TYPE_GIG_DIMENSION_TYPE (ipatch_gig_dimension_type_get_type())
/* enumerations from "IpatchGigEffects.h" */
GType ipatch_gig_filter_type_get_type (void);
#define IPATCH_TYPE_GIG_FILTER_TYPE (ipatch_gig_filter_type_get_type())
GType ipatch_gig_control_type_get_type (void);
#define IPATCH_TYPE_GIG_CONTROL_TYPE (ipatch_gig_control_type_get_type())
/* enumerations from "IpatchGigRegion.h" */
GType ipatch_gig_region_flags_get_type (void);
#define IPATCH_TYPE_GIG_REGION_FLAGS (ipatch_gig_region_flags_get_type())
/* enumerations from "IpatchGigSubRegion.h" */
GType ipatch_gig_sub_region_flags_get_type (void);
#define IPATCH_TYPE_GIG_SUB_REGION_FLAGS (ipatch_gig_sub_region_flags_get_type())
/* enumerations from "IpatchItem.h" */
GType ipatch_item_flags_get_type (void);
#define IPATCH_TYPE_ITEM_FLAGS (ipatch_item_flags_get_type())
/* enumerations from "IpatchPaste.h" */
GType ipatch_paste_choice_get_type (void);
#define IPATCH_TYPE_PASTE_CHOICE (ipatch_paste_choice_get_type())
GType ipatch_paste_priority_get_type (void);
#define IPATCH_TYPE_PASTE_PRIORITY (ipatch_paste_priority_get_type())
/* enumerations from "IpatchRiff.h" */
GType ipatch_riff_status_get_type (void);
#define IPATCH_TYPE_RIFF_STATUS (ipatch_riff_status_get_type())
GType ipatch_riff_mode_get_type (void);
#define IPATCH_TYPE_RIFF_MODE (ipatch_riff_mode_get_type())
GType ipatch_riff_chunk_type_get_type (void);
#define IPATCH_TYPE_RIFF_CHUNK_TYPE (ipatch_riff_chunk_type_get_type())
GType ipatch_riff_error_get_type (void);
#define IPATCH_TYPE_RIFF_ERROR (ipatch_riff_error_get_type())
/* enumerations from "IpatchSF2.h" */
GType ipatch_sf2_flags_get_type (void);
#define IPATCH_TYPE_SF2_FLAGS (ipatch_sf2_flags_get_type())
GType ipatch_sf2_info_type_get_type (void);
#define IPATCH_TYPE_SF2_INFO_TYPE (ipatch_sf2_info_type_get_type())
/* enumerations from "IpatchSF2Gen.h" */
GType ipatch_sf2_gen_props_type_get_type (void);
#define IPATCH_TYPE_SF2_GEN_PROPS_TYPE (ipatch_sf2_gen_props_type_get_type())
GType ipatch_sf2_gen_type_get_type (void);
#define IPATCH_TYPE_SF2_GEN_TYPE (ipatch_sf2_gen_type_get_type())
GType ipatch_sf2_gen_sample_modes_get_type (void);
#define IPATCH_TYPE_SF2_GEN_SAMPLE_MODES (ipatch_sf2_gen_sample_modes_get_type())
/* enumerations from "IpatchSF2Mod.h" */
GType ipatch_sf2_mod_field_masks_get_type (void);
#define IPATCH_TYPE_SF2_MOD_FIELD_MASKS (ipatch_sf2_mod_field_masks_get_type())
GType ipatch_sf2_mod_field_shifts_get_type (void);
#define IPATCH_TYPE_SF2_MOD_FIELD_SHIFTS (ipatch_sf2_mod_field_shifts_get_type())
GType ipatch_sf2_mod_control_get_type (void);
#define IPATCH_TYPE_SF2_MOD_CONTROL (ipatch_sf2_mod_control_get_type())
GType ipatch_sf2_mod_control_palette_get_type (void);
#define IPATCH_TYPE_SF2_MOD_CONTROL_PALETTE (ipatch_sf2_mod_control_palette_get_type())
GType ipatch_sf2_mod_direction_get_type (void);
#define IPATCH_TYPE_SF2_MOD_DIRECTION (ipatch_sf2_mod_direction_get_type())
GType ipatch_sf2_mod_polarity_get_type (void);
#define IPATCH_TYPE_SF2_MOD_POLARITY (ipatch_sf2_mod_polarity_get_type())
GType ipatch_sf2_mod_type_get_type (void);
#define IPATCH_TYPE_SF2_MOD_TYPE (ipatch_sf2_mod_type_get_type())
GType ipatch_sf2_mod_transform_get_type (void);
#define IPATCH_TYPE_SF2_MOD_TRANSFORM (ipatch_sf2_mod_transform_get_type())
GType ipatch_sf2_mod_flags_get_type (void);
#define IPATCH_TYPE_SF2_MOD_FLAGS (ipatch_sf2_mod_flags_get_type())
/* enumerations from "IpatchSF2Sample.h" */
GType ipatch_sf2_sample_channel_get_type (void);
#define IPATCH_TYPE_SF2_SAMPLE_CHANNEL (ipatch_sf2_sample_channel_get_type())
GType ipatch_sf2_sample_flags_get_type (void);
#define IPATCH_TYPE_SF2_SAMPLE_FLAGS (ipatch_sf2_sample_flags_get_type())
/* enumerations from "IpatchSF2VoiceCache.h" */
GType ipatch_sf2_voice_sel_type_get_type (void);
#define IPATCH_TYPE_SF2_VOICE_SEL_TYPE (ipatch_sf2_voice_sel_type_get_type())
/* enumerations from "IpatchSample.h" */
GType ipatch_sample_loop_type_get_type (void);
#define IPATCH_TYPE_SAMPLE_LOOP_TYPE (ipatch_sample_loop_type_get_type())
/* enumerations from "IpatchSampleData.h" */
GType ipatch_sample_data_migrate_flags_get_type (void);
#define IPATCH_TYPE_SAMPLE_DATA_MIGRATE_FLAGS (ipatch_sample_data_migrate_flags_get_type())
/* enumerations from "IpatchSampleStoreRam.h" */
GType ipatch_sample_store_ram_flags_get_type (void);
#define IPATCH_TYPE_SAMPLE_STORE_RAM_FLAGS (ipatch_sample_store_ram_flags_get_type())
/* enumerations from "IpatchSndFile.h" */
GType ipatch_snd_file_endian_get_type (void);
#define IPATCH_TYPE_SND_FILE_ENDIAN (ipatch_snd_file_endian_get_type())
/* enumerations from "IpatchTypeProp.h" */
GType ipatch_splits_type_get_type (void);
#define IPATCH_TYPE_SPLITS_TYPE (ipatch_splits_type_get_type())
/* enumerations from "IpatchUnit.h" */
GType ipatch_unit_flags_get_type (void);
#define IPATCH_TYPE_UNIT_FLAGS (ipatch_unit_flags_get_type())
GType ipatch_unit_type_get_type (void);
#define IPATCH_TYPE_UNIT_TYPE (ipatch_unit_type_get_type())
GType ipatch_unit_class_type_get_type (void);
#define IPATCH_TYPE_UNIT_CLASS_TYPE (ipatch_unit_class_type_get_type())
/* enumerations from "IpatchVBankRegion.h" */
GType ipatch_vbank_region_note_range_mode_get_type (void);
#define IPATCH_TYPE_VBANK_REGION_NOTE_RANGE_MODE (ipatch_vbank_region_note_range_mode_get_type())
GType ipatch_vbank_region_root_note_mode_get_type (void);
#define IPATCH_TYPE_VBANK_REGION_ROOT_NOTE_MODE (ipatch_vbank_region_root_note_mode_get_type())
/* enumerations from "misc.h" */
GType ipatch_error_get_type (void);
#define IPATCH_TYPE_ERROR (ipatch_error_get_type())
/* enumerations from "sample.h" */
GType ipatch_sample_width_get_type (void);
#define IPATCH_TYPE_SAMPLE_WIDTH (ipatch_sample_width_get_type())
GType ipatch_sample_channel_get_type (void);
#define IPATCH_TYPE_SAMPLE_CHANNEL (ipatch_sample_channel_get_type())
GType ipatch_sample_channel_type_get_type (void);
#define IPATCH_TYPE_SAMPLE_CHANNEL_TYPE (ipatch_sample_channel_type_get_type())
GType ipatch_sample_sign_get_type (void);
#define IPATCH_TYPE_SAMPLE_SIGN (ipatch_sample_sign_get_type())
GType ipatch_sample_endian_get_type (void);
#define IPATCH_TYPE_SAMPLE_ENDIAN (ipatch_sample_endian_get_type())
G_END_DECLS

#endif /* __IPATCH_BUILTIN_ENUMS_H__ */

/* Generated data ends here */

